import pytest

from lailib.image.augment import *


class TestJPGCompress:
    def test_acc_mc(self):
        img_rgb = np.array(np.random.rand(802, 604, 3) * 255, dtype=np.uint8)
        img_re = jpg_compress(img_rgb)
        assert img_re.shape == img_rgb.shape

    def test_acc_m(self):
        img_rgb = np.array(np.random.rand(128, 256) * 255, dtype=np.uint8)
        img_re = jpg_compress(img_rgb)
        assert img_re.shape == img_rgb.shape

    def test_acc_xlc(self):
        img_rgb = np.array(np.random.rand(4000, 4000, 3) * 255, dtype=np.uint8)
        img_re = jpg_compress(img_rgb)
        assert img_re.shape == img_rgb.shape

    def test_acc_xsc(self):
        img_rgb = np.array(np.random.rand(2, 2, 3) * 255, dtype=np.uint8)
        img_re = jpg_compress(img_rgb)
        assert img_re.shape == img_rgb.shape

    def test_acc_xl(self):
        img_rgb = np.array(np.random.rand(4000, 4000) * 255, dtype=np.uint8)
        img_re = jpg_compress(img_rgb)
        assert img_re.shape == img_rgb.shape

    def test_acc_xs(self):
        img_rgb = np.array(np.random.rand(2, 2) * 255, dtype=np.uint8)
        img_re = jpg_compress(img_rgb)
        assert img_re.shape == img_rgb.shape

    def test_acc_low_qual(self):
        img_rgb = np.array(np.random.rand(800, 458, 3) * 255, dtype=np.uint8)
        img_re = jpg_compress(img_rgb, 0)
        assert img_re.shape == img_rgb.shape

    def test_acc_high_qual(self):
        img_rgb = np.array(np.random.rand(128, 256, 3) * 255, dtype=np.uint8)
        img_re = jpg_compress(img_rgb, 100)
        assert img_re.shape == img_rgb.shape

    def test_float_c(self):
        img_rgb = np.random.rand(128, 256, 3)
        with pytest.raises(ValueError):
            img_re = jpg_compress(img_rgb)

    def test_float(self):
        img_rgb = np.random.rand(128, 256)
        with pytest.raises(ValueError):
            img_re = jpg_compress(img_rgb)



import os
import pytest
import cv2
import numpy as np

from lailib.image.convert import *


class TestBase64:
    def test_acc_base64_color(self, tmpdir):
        img_rgb = np.array(np.random.rand(128, 256, 3) * 255, dtype=np.uint8)
        img_path = os.path.join(tmpdir, "test_color.png")
        cv2.imwrite(img_path, img_rgb)
        base64_str = base64_from_file(img_path)
        img_from_str = base64_to_img(base64_str)
        assert img_from_str.shape == img_rgb.shape
        assert np.array_equal(img_from_str, img_rgb)

    def test_acc_base64_gray(self, tmpdir):
        img = np.array(np.random.rand(128, 256) * 255, dtype=np.uint8)
        img_path = os.path.join(tmpdir, "test_color.png")
        cv2.imwrite(img_path, img)
        base64_str = base64_from_file(img_path)
        img_from_str = base64_to_img(base64_str, rgb=False)
        assert img_from_str.shape == img.shape
        assert np.array_equal(img_from_str, img)

    def test_acc_base64_color_xl(self, tmpdir):
        img_rgb = np.array(np.random.rand(4000, 4000, 3) * 255, dtype=np.uint8)
        img_path = os.path.join(tmpdir, "test_color.png")
        cv2.imwrite(img_path, img_rgb)
        base64_str = base64_from_file(img_path)
        img_from_str = base64_to_img(base64_str)
        assert img_from_str.shape == img_rgb.shape
        assert np.array_equal(img_from_str, img_rgb)

    def test_acc_base64_gray_xl(self, tmpdir):
        img = np.array(np.random.rand(4000, 4000) * 255, dtype=np.uint8)
        img_path = os.path.join(tmpdir, "test_color.png")
        cv2.imwrite(img_path, img)
        base64_str = base64_from_file(img_path)
        img_from_str = base64_to_img(base64_str, rgb=False)
        assert img_from_str.shape == img.shape
        assert np.array_equal(img_from_str, img)

    def test_acc_base64_color_xs(self, tmpdir):
        img_rgb = np.array(np.random.rand(2, 2, 3) * 255, dtype=np.uint8)
        img_path = os.path.join(tmpdir, "test_color.png")
        cv2.imwrite(img_path, img_rgb)
        base64_str = base64_from_file(img_path)
        img_from_str = base64_to_img(base64_str)
        assert img_from_str.shape == img_rgb.shape
        assert np.array_equal(img_from_str, img_rgb)

    def test_acc_base64_gray_xs(self, tmpdir):
        img = np.array(np.random.rand(2, 2) * 255, dtype=np.uint8)
        img_path = os.path.join(tmpdir, "test_color.png")
        cv2.imwrite(img_path, img)
        base64_str = base64_from_file(img_path)
        img_from_str = base64_to_img(base64_str, rgb=False)
        assert img_from_str.shape == img.shape
        assert np.array_equal(img_from_str, img)


class TestSplit:
    def test_acc_mc(self):
        img = np.array(np.random.rand(1280, 2560, 3) * 255, dtype=np.uint8)
        metas = split_square_img_boarder(img)
        re_img = combine_square_img_boarder(metas)
        assert np.array_equal(img, re_img)

    def test_acc_mc2(self):
        img = np.array(np.random.rand(1280, 2560) * 255, dtype=np.uint8)
        metas = split_square_img_boarder(img)
        re_img = combine_square_img_boarder(metas)

        assert np.array_equal(img, re_img)

    def test_acc_mc3(self):
        img = np.array(np.random.rand(300, 300, 3) * 255, dtype=np.uint8)
        metas = split_square_img_boarder(img)
        re_img = combine_square_img_boarder(metas)

        assert np.array_equal(img, re_img)

    def test_acc_mc4(self):
        img = np.array(np.random.rand(512, 512) * 255, dtype=np.uint8)
        metas = split_square_img_boarder(img)
        re_img = combine_square_img_boarder(metas)

        assert np.array_equal(img, re_img)

    def test_s1(self):
        img = np.array(np.random.rand(256, 256, 3) * 255, dtype=np.uint8)
        metas = split_square_img_boarder(img)
        re_img = combine_square_img_boarder(metas)

        assert np.array_equal(img, re_img)

    def test_s3(self):
        img = np.array(np.random.rand(256, 256) * 255, dtype=np.uint8)
        metas = split_square_img_boarder(img)
        re_img = combine_square_img_boarder(metas)

        assert np.array_equal(img, re_img)

    def test_s2(self):
        img = np.array(np.random.rand(135, 142, 3) * 255, dtype=np.uint8)
        with pytest.raises(ValueError):
            metas = split_square_img_boarder(img)

    def test_s4(self):
        img = np.array(np.random.rand(13, 12) * 255, dtype=np.uint8)
        with pytest.raises(ValueError):
            metas = split_square_img_boarder(img)

    def test_s5(self):
        img = np.array(np.random.rand(257, 257) * 255, dtype=np.uint8)
        metas = split_square_img_boarder(img)
        re_img = combine_square_img_boarder(metas)
        assert np.array_equal(img, re_img)

    def test_s5(self):
        img = np.array(np.random.rand(257, 257) * 255, dtype=np.uint8)
        metas = split_square_img_boarder(img)
        re_img = combine_square_img_boarder(metas)
        assert np.array_equal(img, re_img)

    def test_s5(self):
        img = np.array(np.random.rand(257, 257) * 255, dtype=np.uint8)
        metas = split_square_img_boarder(img,h=10,w=10,boarder=1)
        re_img = combine_square_img_boarder(metas,boarder=1)
        assert np.array_equal(img, re_img)







import os
from os import listdir

import numpy as np
import torch


# TODO add support for optimizer tensor type
# TODO add types for function heads
# TODO support early stopping
def save_network(network,
                 optimizer,
                 save_dir,
                 model_name,
                 global_step,
                 use_gpu=True):
    '''
    save current neural network parameters and optimizer parameters
    currently the use_gpu flag only takes care of parameters in network,
    some optimizers has tensors in it. Be careful on optimizers when use
    the save load functions

    :param network: the pytorch neural network that should be saved
    :param optimizer: the optimizer for network training
    :param save_dir: path to the checkpoint directory
    :param model_name: name of the model
    :param global_step: An integer indicates how many steps the training
                        had run, (usually steps is the number of iterations and
                        number of epoches)
    :param use_gpu: a parameter indicates if the input neural network is on gpu
    :return: None
    '''
    if '_' in model_name or '.' in model_name:
        raise ValueError('model name can not contain "." or "_"')
    save_filename = '%s_%s.pth' % (model_name, global_step)
    save_path = os.path.join(save_dir, save_filename)
    state = {'state_dict': network.cpu().state_dict(),
             'optimizer': optimizer.state_dict()}
    torch.save(state, save_path)
    if use_gpu:
        network.cuda()


# TODO add support for optimizer tensor type

def load_last_checkpoint(network,
                         optimizer,
                         save_dir,
                         model_name,
                         use_gpu,
                         reset_optimizer=False):
    '''
    load the latest checkpoint from checkpoint folder
    currently the use_gpu flag only takes cares of parameters in network,
    some optimizers has tensors in it. Be careful on optimizers when use
    the save load functions
    :param network: the pytorch neural network
    :param optimizer: the optimizer for network training
    :param save_dir: where the checkpoint file sits
    :param model_name: name of the model
    :param use_gpu: a parameter indicates if the output neural network is on gpu
    :reset_optimizer: if set to true, optimizer will not load parameters in checkpoint dict
    :return: network with loaded parameters, optimizer with loaded parameter, iternumber of last checkpoint
    '''
    if '_' in model_name or '.' in model_name:
        raise ValueError('model name can not contain "." or "_"')
    checkpoint_paths = [f for f in listdir(save_dir) if f.endswith('pth')]
    if not checkpoint_paths:
        print('first iteration, initialize model')
        return network, 0
    raw_names = [f.split('.')[0] for f in checkpoint_paths]
    iter_numbers = [int(f.split('_')[1]) for f in raw_names]
    max_global_step = np.max(np.asarray(iter_numbers))
    network, optimizer = load_network(network,
                                      optimizer,
                                      save_dir,
                                      model_name,
                                      max_global_step,
                                      use_gpu,
                                      reset_optimizer)
    return network, optimizer, max_global_step


def load_network(network,
                 optimizer,
                 save_dir,
                 model_name,
                 global_step,
                 use_gpu,
                 reset_optimizer=False):
    '''
    load neural network parameters
    currently the use_gpu flag only takes cares of parameters in network,
    some optimizers has tensors in it. Be careful on optimizers when use
    the save load functions
    :param network: the pytorch neural network (to be safe the model should be on cpu)
    :param optimizer: the optimizer for network training
    :param save_dir: where the checkpoint file sits
    :param model_name: name of the model
    :param global_step: An integer indicates how many steps the training
                        had run, (usually steps is the number of iterations and
                        number of epoches)
    :param reset_optimizer: if set to true, optimizer will not load parameters in checkpoint dict

    :return: network with loaded parameters, optimizer with loaded parameter,
    '''
    if '_' in model_name or '.' in model_name:
        raise ValueError('model name can not contain "." or "_"')
    network.cpu()
    save_filename = '%s_%s.pth' % (model_name, global_step)
    save_path = os.path.join(save_dir, save_filename)
    state_dicts = torch.load(save_path)
    network.load_state_dict(state_dicts['state_dict'])
    if use_gpu:
        network.cuda()
    if not reset_optimizer:
        optimizer.load_state_dict(state_dicts['optimizer'])
    print('load checkpoint from {}'.format(save_filename))
    return network, optimizer



from os import listdir

import pytest
import torch
import torch.nn as nn
from torch.nn import init

from lailib.torch.parameter_store import save_network, load_last_checkpoint, load_network


# TODO add test for optimizer
# TODO test reset parameter

@pytest.fixture(scope='function')
def model_and_meta():
    model_name = 'Dummy'
    global_step = 0
    saved_model = DummyTorchModule()
    return [model_name, global_step, saved_model]


class DummyTorchModule(nn.Module):
    def __init__(self):
        super(DummyTorchModule, self).__init__()
        self.test_weight = torch.nn.Parameter(init.orthogonal_(torch.Tensor(5, 5)))


class TestSaveLoad():
    @staticmethod
    def compare_models(model_1, model_2):
        for key_item_1, key_item_2 in zip(model_1.state_dict().items(), model_2.state_dict().items()):
            if not torch.equal(key_item_1[1], key_item_2[1]):
                if (key_item_1[0] == key_item_2[0]):
                    return False
                else:
                    raise Exception
        return True

    def test_save_model(self, tmpdir, model_and_meta):
        model_name, global_step, saved_model = model_and_meta
        saved_optimizer = torch.optim.Adagrad(saved_model.parameters(), lr=5e-4)
        save_network(saved_model,
                     saved_optimizer,
                     tmpdir,
                     model_name,
                     global_step,
                     use_gpu=False)
        assert ('%s_%s.pth' % (model_name, global_step) in listdir(tmpdir))

    # def test_save_model_gpu(self, tmpdir, model_and_meta):
    #     model_name, global_step, saved_model = model_and_meta
    #     saved_model = DummyTorchModule().cuda()
    #     optimizer = torch.optim.Adagrad(saved_model.parameters(), lr=5e-4)
    #     save_network(saved_model,
    #                  optimizer,
    #                  tmpdir,
    #                  model_name,
    #                  global_step)
    #
    #     assert ('%s_%s.pth' % (model_name, global_step) in listdir(tmpdir))

    def test_save_load(self, tmpdir, model_and_meta):
        # test if load saved model we get same result
        model_name, global_step, saved_model = model_and_meta
        saved_optimizer = torch.optim.Adagrad(saved_model.parameters(), lr=5e-4)

        save_network(saved_model,
                     saved_optimizer,
                     tmpdir,
                     model_name,
                     global_step,
                     use_gpu=False, )
        reload_model = DummyTorchModule()
        reload_optimizer = torch.optim.Adagrad(reload_model.parameters(), lr=5e-4)
        assert (not self.compare_models(saved_model, reload_model))
        reload_model, reload_optimizer = load_network(reload_model,
                                                      reload_optimizer,
                                                      tmpdir,
                                                      model_name,
                                                      global_step,
                                                      use_gpu=False,
                                                      reset_optimizer=False)
        assert (self.compare_models(saved_model, reload_model))

    # def test_save_load_gpu(self, tmpdir, model_and_meta):
    #     # test if load saved model we get same result
    #     model_name, global_step, saved_model = model_and_meta
    #     saved_model = saved_model.cuda()
    #     saved_optimizer = torch.optim.Adagrad(saved_model.parameters(), lr=5e-4)
    #
    #     reload_model = DummyTorchModule().cuda()
    #     reload_optimizer = torch.optim.Adagrad(reload_model.parameters(), lr=5e-4)
    #
    #     save_network(saved_model,
    #                  saved_optimizer,
    #                  tmpdir,
    #                  model_name,
    #                  global_step,
    #                  use_gpu=True)
    #
    #     # it's close to impossible that two initialized parameter matrix is exactly the same
    #     assert (not self.compare_models(saved_model, reload_model))
    #     reload_model, reload_optimizer = load_network(reload_model,
    #                                                   reload_optimizer,
    #                                                   tmpdir,
    #                                                   model_name,
    #                                                   global_step,
    #                                                   use_gpu=True,
    #                                                   reset_optimizer=False)
    #     # test if reload model is the same as test model
    #     assert (self.compare_models(saved_model, reload_model))

    def test_save_load_last_checkpoint(self, tmpdir):
        model_name = 'Dummy'
        global_step_0 = 0
        saved_model_0 = DummyTorchModule()
        saved_optimizer_0 = torch.optim.Adagrad(saved_model_0.parameters(), lr=5e-4)

        global_step_1 = 1
        saved_model_1 = DummyTorchModule()
        saved_optimizer_1 = torch.optim.Adagrad(saved_model_1.parameters(), lr=5e-4)

        reload_model = DummyTorchModule()
        reload_optimizer = torch.optim.Adagrad(reload_model.parameters(), lr=5e-4)
        save_network(saved_model_0,
                     saved_optimizer_0,
                     tmpdir,
                     model_name,
                     global_step_0,
                     use_gpu=False)

        save_network(saved_model_1,
                     saved_optimizer_1,
                     tmpdir,
                     model_name,
                     global_step_1,
                     use_gpu=False)

        reload_model, reload_optimizer, global_step = load_last_checkpoint(reload_model,
                                                                           reload_optimizer,
                                                                           tmpdir,
                                                                           model_name,
                                                                           use_gpu=False)
        assert (global_step == 1)
        assert (not self.compare_models(saved_model_0, reload_model))
        assert (self.compare_models(saved_model_1, reload_model))

    def test_invalid_model_name(self, model_and_meta, tmpdir):
        model_name, global_step, saved_model = model_and_meta
        model_name = 'dummy_model'
        saved_optimizer = torch.optim.Adagrad(saved_model.parameters(), lr=5e-4)

        with pytest.raises(ValueError, match='model name can not contain "." or "_"'):
            save_network(saved_model,
                         saved_optimizer,
                         tmpdir,
                         model_name,
                         global_step=False,
                         use_gpu=True)

        with pytest.raises(ValueError, match='model name can not contain "." or "_"'):
            load_last_checkpoint(saved_model,
                                 saved_optimizer,
                                 tmpdir,
                                 model_name,
                                 use_gpu=False,
                                 reset_optimizer=False)

        with pytest.raises(ValueError, match='model name can not contain "." or "_"'):
            load_network(saved_model,
                         saved_optimizer,
                         tmpdir,
                         model_name,
                         global_step,
                         use_gpu=False,
                         reset_optimizer=False)



import cv2
import matplotlib.pyplot as plt


def show_img_np(np_array, size=20, save=False):
    """
    :param np_array: input image, one channel or 3 channel,
    :param save: if save image
    :param size: image canvas size, in inches
    :return:
    """
    if len(np_array.shape) < 3:
        plt.rcParams['image.cmap'] = 'gray'
    img_np = np_array
    plt.figure(figsize=(size, size), facecolor='w', edgecolor='k')
    plt.imshow(img_np)
    if save:
        cv2.imwrite('debug.png', img_np)
    else:
        plt.show()



import cv2


def otsu_thresh(im):
    '''
    binarize image with otsu algorithm
    :param im(ndarray): uint8 numpy array
    :return: binarized image as uint8 numpy array
    '''
    _, binarized = cv2.threshold(im, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    return binarized


def vanilla_thresh(im, th=128):
    '''
    binarize image with given threshold(default is 128)
    :param im(ndarray): uint8 numpy array
    :param th(int): threshold
    :return: binarized image as uint8 numpy array
    '''
    _, binarized = cv2.threshold(im, 0, th, cv2.THRESH_BINARY)
    return binarized



import numpy as np
import cv2
from lailib.image.binarize import otsu_thresh
import random


def random_sample(imgs, target_h=256, target_w=256, size_aug_min=0.25, size_aug_max=4, skew_aug=2,
                  randomizer='uniform_exp'):
    """
    random sample parts from images with the same size
    :param img: original image
    :param target_h: output h
    :param target_w: output w
    :param size_aug_min: min zoom scale when sampling image (sample bigger parts and zoom out)
    :param size_aug_max: max zoom scale when sampling image (sample smaller parts and zoom in)
    :param skew_aug: scale for random stretching image when sampling (augmentation)
    :param randomizer: randomize method when randomizing sample scale
    :return: sampled image
    """
    size_aug_min,size_aug_max = 1/size_aug_max,1/size_aug_min
    img = imgs[0]
    if len(img.shape) == 3:
        if_color = True
        h, w, _ = img.shape
    else:
        h, w = img.shape
        if_color = False
    if randomizer == 'uniform':
        size_aug = random.uniform(1 / size_aug_min, 1 / size_aug_max)
    elif randomizer == 'uniform_exp':
        size_aug = np.exp(random.uniform(np.log(1 / size_aug_min), np.log(1 / size_aug_max)))
    else:
        raise ValueError('randomizer not supported')
    if size_aug_min*min(h,w)<1:
        raise ValueError('lower size limit is too small')

    skew_aug = np.exp(random.uniform(-np.log(skew_aug), np.log(skew_aug)))
    size_aug = np.min([size_aug, h / target_h / np.sqrt(skew_aug), w / target_w * np.sqrt(skew_aug)])
    crop_h, crop_w = int(target_h * size_aug * np.sqrt(skew_aug)), int(target_w * size_aug / np.sqrt(skew_aug))
    crop_h,crop_w = max(crop_h,1),max(crop_w,1)
    #     print(crop_h,crop_w)
    min_y, max_y, min_x, max_x = 0, h - crop_h, 0, w - crop_w

    y_start = random.randint(min_y, max_y + 1)
    x_start = random.randint(min_x, max_x + 1)
    return_imgs = []
    for img in imgs:
        if if_color:
            raw_crop = img[y_start:y_start + crop_h, x_start:x_start + crop_w, :]
        else:
            raw_crop = img[y_start:y_start + crop_h, x_start:x_start + crop_w]
        if target_h < crop_h:
            return_imgs.append(cv2.resize(raw_crop, (target_h, target_w), interpolation=cv2.INTER_CUBIC))
        else:
            return_imgs.append(cv2.resize(raw_crop, (target_h, target_w), interpolation=cv2.INTER_AREA))
    return return_imgs



def crop_boundary_and_padding(im, padding=0, binarized=None):
    '''
    crop and padding objects and text imgs with pure black border, then padding.
    ex:
    for input
        0   255 0   0
        0   255 255 0
        0   255 0   0
        0   0   0   0
    output is
        255 0
        255 255
        255 0
    :param image_cv(ndarray): input image, uint8 numpy array, assumed to be gray scale
    :param padding(int or list of ints): if padding is a scalar, padding will be applied to left, right, upside and
    downside of an image.
                    if padding is a list of size 4, each entry in the list corresponds to left, right, upside and
                    downside of an image.
    :return: cropped image (uint8)
    '''
    if len(im.shape) != 2 or im.dtype != np.uint8:
        raise TypeError('input image must be gray scale image as uint8 ndarray')
    if not binarized is None:
        if binarized.shape != im.shape:
            raise ValueError('binarized image shape {} must meet in image shape {}'.format(binarized.shape, im.shape))
        if binarized.dtype != np.uint8:
            raise TypeError('binarized mask must be uint8 ndarray')
    else:
        binarized = otsu_thresh(im)
    if np.sum(binarized) == 0:
        raise ValueError('In image for crop function is all zero')

    if isinstance(padding, int):
        padding = [padding] * 4
    elif not (isinstance(padding, list) and len(padding) == 4):
        raise TypeError('padding in crop function must be int or list of size 4')

    # use binarized image to get vertical and horizontal boundaries
    col_sums = np.sum(binarized, axis=1)
    row_start = np.where(col_sums)[0][0]
    rev_col_sum = np.flip(col_sums, axis=0)
    rev_end = np.where(rev_col_sum)[0][0]
    row_end = col_sums.shape[0] - rev_end

    row_sums = np.sum(binarized, axis=0)
    col_start = np.where(row_sums)[0][0]
    rev_row_sum = np.flip(row_sums, axis=0)
    rev_end = np.where(rev_row_sum)[0][0]
    col_end = row_sums.shape[0] - rev_end

    cropped = im[row_start:row_end, col_start:col_end]

    cropped_height, cropped_width = cropped.shape
    # TODO add rand noise to height
    final_out = np.zeros((cropped_height + padding[2] + padding[3],
                          cropped_width + padding[0] + padding[1]), dtype=np.uint8)

    out_height, out_width = final_out.shape
    # do not directly use -padding as the value for end side because it can't deal with padding = 0
    final_out[padding[2]: out_height - padding[3], padding[0]: out_width - padding[1]] = cropped
    return final_out





import cv2


def resize_height_keep_ratio(im, new_height, **kwargs):
    '''
    resize image with given height, keep ratio, kwargs are passed to
    cv2.resize function.
    :param image_cv(ndarray): image to be resized
    :param new_height(int): new height of the output image
    :return: resized image
    '''
    (height, width) = im.shape[:2]
    new_width = int(float(width) * new_height / float(height))
    im = cv2.resize(im, (new_width, new_height), **kwargs)
    return im



import pytest

from lailib.image.check import *


class TestCheckImgFormat:
    def test_acc_color(self):
        img = np.array(np.random.rand(128, 256, 3) * 255, dtype=np.uint8)
        check_img_format(img, isrgb=True)
        with pytest.raises(ValueError):
            check_img_format(img, isrgb=False)

    def test_acc_gray(self):
        img = np.array(np.random.rand(128, 256) * 255, dtype=np.uint8)
        check_img_format(img)
        with pytest.raises(ValueError):
            check_img_format(img, isrgb=True)

    def test_acc_multichannel(self):
        img = np.array(np.random.rand(4000, 4000, 4) * 255, dtype=np.uint8)
        with pytest.raises(ValueError):
            check_img_format(img, isrgb=False)

    def test_acc_one_channel(self):
        img = np.array(np.random.rand(256) * 255, dtype=np.uint8)
        with pytest.raises(ValueError):
            check_img_format(img, isrgb=True)
        with pytest.raises(ValueError):
            check_img_format(img, isrgb=False)



import base64
from collections import namedtuple
import cv2
import numpy as np


def base64_to_img(str, rgb=True):
    """
    convert base64 string to numpy image
    :param str: (string)base64 encoded string of image
    :param rgb: (Bool)if decoded image is rgb,
    :return:
    """
    mode = 0
    if rgb:
        mode = 1
    img = base64.b64decode(str)
    img_np = np.frombuffer(img, np.uint8)
    img = cv2.imdecode(img_np, mode)
    return img


def base64_from_file(path):
    """
    get base64 image string from file path
    :param path: (string)path to image file
    :return: (string)base64 image string
    """
    with open(path, 'rb') as handler:
        return base64.b64encode(handler.read()).decode('utf-8')


def split_square_img_boarder(img, h=256, w=256, boarder=25):
    """
    split image into a bag of small images for batch training
    :param img: input image
    :param h: height of bag of small image
    :param w: width of bag of small image
    :param boarder: to eliminate artifact, how far boarder of small image to discard
    :return: list of namedtuple of (y:y_coord of lr corner of the small image,
                                    x:x_coord of lr corner of the small image,
                                    img:small image)
    """
    b=boarder*2
    img_h, img_w = img.shape[0], img.shape[1]
    if img_h < h or img_w < w:
        raise ValueError("input image should be larger than split img")
    if h<b:
        raise ValueError("boarder is too large")
    h_start = 0
    w_start = 0
    h_start_list = []
    w_start_list = []
    while h_start + h < img_h:
        h_start_list.append(h_start)
        h_start += (h - b)
    while w_start + h < img_w:
        w_start_list.append(w_start)
        w_start += (w - b)
    h_start_list.append(img_h - h)
    w_start_list.append(img_w - w)
    #     print(h_start_list)
    #     print(w_start_list)
    meta = namedtuple('meta',['y','x','img'])
    split_list = []
    for i in h_start_list:
        for j in w_start_list:
            if len(img.shape) == 3:
                split_list.append(meta(i, j, img[i:i + h, j:j + w, :]))
            else:
                split_list.append(meta(i, j, img[i:i + h, j:j + w]))
    return split_list


def combine_square_img_boarder(metas, boarder=25):
    """
    combine bag of small image back into original one
    :param metas: bag of small image,
                  list of namedtuple of (y:y_coord of lr corner of the small image,
                                         x:x_coord of lr corner of the small image,
                                         img:small image)
    :param boarder: to eliminate artifact, how far boarder of small image to discard,
    should be the same as split function
    :return: original image
    """
    square_h = metas[0].img.shape[0]
    square_w = metas[0].img.shape[1]
    img_h = max(x.y for x in metas) + square_h
    img_w = max(x.x for x in metas) + square_w

    #     print(img_h,img_w)
    if len(metas[0].img.shape) == 3:
        blank_img = np.zeros((img_h, img_w,3))

        for i in metas:
            blank_img[i.y:i.y + square_h, i.x:i.x + square_w,:] = i.img
        for i in metas:
            w_boarder, s_boarder, a_boarder, d_boarder = boarder, boarder, boarder, boarder
            blank_img[i.y + boarder:i.y + square_h - boarder,
            i.x + boarder:i.x + square_w - boarder,:] = i.img[boarder:-boarder, boarder:-boarder,:]


    else:
        blank_img = np.zeros((img_h, img_w))
        for i in metas:
            blank_img[i.y:i.y + square_h, i.x:i.x + square_w] = i.img
        for i in metas:
            w_boarder, s_boarder, a_boarder, d_boarder = boarder, boarder, boarder, boarder
            blank_img[i.y + boarder:i.y + square_h - boarder,
                      i.x + boarder:i.x + square_w - boarder] = i.img[boarder:-boarder,boarder:-boarder]
    return blank_img


import numpy as np
import pytest

from lailib.image.resize import resize_height_keep_ratio


@pytest.mark.parametrize('original_height,original_width,goal_height,goal_width',
                         [(1, 1, 100, 100), (1, 1, 30, 30), (3, 5, 50, 83), (5, 3, 50, 30), (100, 250, 30, 75)])
def test_resize_height_keep_ratio(original_height, original_width, goal_height, goal_width):
    fake_im = np.zeros((original_height, original_width))
    ret_im = resize_height_keep_ratio(fake_im, goal_height)
    _, w = ret_im.shape
    assert (abs(goal_width - w) < 2)



from lailib.misc.misc import *

class TestRandomStr:
    def test_len(self):
        gened_str = set([])
        for i in range(100):
            rand_str = gen_random_str()
            assert rand_str not in gened_str
            gened_str.add(rand_str)

    def test_heck_len(self):
        for i in range(100):
            str_len = random.randint(1,10)
            rand_str = gen_random_str(str_len)
            assert len(rand_str)==str_len



import numpy as np


def check_img_format(img, check_color_space=True, isrgb=False):
    """
    check image format, must be uint8 numpy array
    :param img: (numpy array)input image
    :param check_color_space: (bool) do checking on color space, if False, both gray and rgb image are supported
    :param isrgb: (boolean)check if image is colored, True for must colored, False for must be gray.
    :return: None
    """
    s = img.shape
    if check_color_space:
        if isrgb and len(s) != 3:
            raise ValueError("Image should be colored, got gray")
        if (not isrgb) and len(s) != 2:
            raise ValueError("Image should be gray, got 3D array with shape of {}".format(str(s)))
    try:
        if img.dtype != np.array([1], dtype=np.uint8).dtype:
            raise ValueError("Image should be numpy array with data type of uint8")
    except Exception:
        raise ValueError("Image should be numpy array with data type of uint8")



import cv2
import numpy as np

from lailib.image.check import check_img_format


def jpg_compress(img, quality=30):
    """
    add jpg compress noise to image
    :param img: (numpy uint8)input image
    :param quality: (int)0-100, jpeg compress quality, lower is lower quality
    :return: (numpy uint8)compressed image
    """
    check_img_format(img, check_color_space=False)
    grey = len(img.shape) == 2
    encode_param = [int(cv2.IMWRITE_JPEG_QUALITY), quality]
    _, result = cv2.imencode('.jpg', np.uint8(img), encode_param)
    if grey:
        result = cv2.imdecode(result, 0)
    else:
        result = cv2.imdecode(result, 1)
    return result



import pytest

from lailib.image.crop import *


class TestCropBoundaryAndPad:
    @staticmethod
    def binarized_case0():
        res = np.zeros((3, 3), dtype=np.uint8)
        res[0, 0] = 255
        res[2, 2] = 255
        padded_in = np.pad(res, ((3, 2), (2, 3)), constant_values=(0, 0), mode='constant')
        return padded_in, res

    @staticmethod
    def binarized_case1():
        res = np.zeros((4, 5), dtype=np.uint8)
        res[0, 0] = 255
        res[3, 4] = 255
        res[2, 1] = 255
        padded_in = np.pad(res, ((23, 2), (2, 44)), constant_values=(0, 0), mode='constant')
        return padded_in, res

    @staticmethod
    def gray_case0():
        '''not binarized image without mask generated outside the crop function'''
        res = np.zeros((4, 5), dtype=np.uint8)
        res[0, 0] = 128
        res[3, 4] = 255
        res[2, 1] = 128
        padded_in = np.pad(res, ((23, 2), (2, 44)), constant_values=(0, 0), mode='constant')
        return padded_in, res

    @staticmethod
    def gray_case1():
        '''gray image with mask generated outside the crop function'''
        res = np.zeros((4, 5), dtype=np.uint8)
        res[0, 0] = 128
        res[3, 4] = 255
        res[2, 1] = 64

        binarized = np.copy(res)
        binarized[0, 0] = 255
        binarized[3, 4] = 255
        binarized[2, 1] = 255
        padded_in = np.pad(res, ((23, 2), (2, 44)), constant_values=(0, 0), mode='constant')
        padded_binarized = np.pad(binarized, ((23, 2), (2, 44)), constant_values=(0, 0), mode='constant')
        return padded_in, padded_binarized, res

    def test_out_type(self):
        x = np.ones((3, 3), dtype=np.uint8)
        out = crop_boundary_and_padding(x)
        assert out.dtype == np.uint8

    def test_binarized(self):
        padded_in0, res_0 = self.binarized_case0()
        padded_in1, res_1 = self.binarized_case1()
        assert (np.array_equal(crop_boundary_and_padding(padded_in0), res_0))
        assert (np.array_equal(crop_boundary_and_padding(padded_in1), res_1))

    def test_gray(self):
        padded_in0, res_0 = self.gray_case0()
        padded_in1, padded_binarized, res_1 = self.gray_case1()
        assert (np.array_equal(crop_boundary_and_padding(padded_in0), res_0))  # without binarize input
        assert (np.array_equal(crop_boundary_and_padding(padded_in1, binarized=padded_binarized),
                               res_1))  # with binarize map

    def test_with_pad(self):
        padded_in_0 = np.ones((3, 3), dtype=np.uint8)
        res_0 = np.zeros((10, 20), dtype=np.uint8)
        res_0[2:5, 10:13] = padded_in_0

        padded_in_1 = np.ones((3, 3), dtype=np.uint8)
        res_1 = np.zeros((9, 9), dtype=np.uint8)
        res_1[3:6, 3:6] = padded_in_1

        assert (np.array_equal(crop_boundary_and_padding(padded_in_0, [10, 7, 2, 5]), res_0))
        assert (np.array_equal(crop_boundary_and_padding(padded_in_1, 3), res_1))

    def test_all_zero_im(self):
        in_im = np.zeros((100, 200), dtype=np.uint8)  # All zero input image with whatever shape
        with pytest.raises(ValueError, match='In image for crop function is all zero'):
            crop_boundary_and_padding(in_im)

    def test_invalid_pad(self):
        wrong_pad0 = [0, 0, 0, 0, 0]
        wrong_pad1 = 15432514.315345
        wrong_pad2 = [0]
        in_im = np.zeros((100, 200), dtype=np.uint8)
        in_im[30, 30] = 255
        with pytest.raises(TypeError, match='padding in crop function must be int or list of size 4'):
            crop_boundary_and_padding(in_im, padding=wrong_pad0)
        with pytest.raises(TypeError, match='padding in crop function must be int or list of size 4'):
            crop_boundary_and_padding(in_im, padding=wrong_pad1)
        with pytest.raises(TypeError, match='padding in crop function must be int or list of size 4'):
            crop_boundary_and_padding(in_im, padding=wrong_pad2)

    def test_wrong_input_type(self):
        wrong_dtype = np.ones((3, 3)) * 255  # image with wrong dtype(int64)
        not_gray_scale = np.ones((5, 5, 3), dtype=np.uint8) * 255  # image with 3 channels
        with pytest.raises(TypeError, match='input image must be gray scale image as uint8 ndarray'):
            crop_boundary_and_padding(wrong_dtype)
        with pytest.raises(TypeError, match='input image must be gray scale image as uint8 ndarray'):
            crop_boundary_and_padding(not_gray_scale)

    def test_wrong_binarized_input(self):
        in_im = np.ones((3, 3), dtype=np.uint8)
        big_mask = np.ones((4, 4), dtype=np.uint8) * 255
        small_mask = np.ones((2, 2), dtype=np.uint8) * 255
        wrong_type = np.ones((3, 3)) * 255

        with pytest.raises(ValueError, match='binarized image shape .* must meet in image shape .*'):
            crop_boundary_and_padding(in_im, binarized=big_mask)
        with pytest.raises(ValueError, match='binarized image shape .* must meet in image shape .*'):
            crop_boundary_and_padding(in_im, binarized=small_mask)
        with pytest.raises(TypeError, match='binarized mask must be uint8 ndarray'):
            crop_boundary_and_padding(in_im, binarized=wrong_type)





from lailib.image.augment import *
from lailib.image.binarize import *
from lailib.image.check import *
from lailib.image.convert import *
from lailib.image.resize import *
from lailib.image.visualize import *
